#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

// limity na liczbe pozycji i ruchow
#define maxN 5000
#define maxM 100000

// liczba pozycji
int N;

// liczba ruchow
int M;

// wlasciciel i ranga kazdej pozycji
int czyje[maxN], ranga[maxN];

// ruchy
int skad[maxM], dokad[maxM];

// funkcja czytajaca liczbe z std wyjscia lub od gracza
int liczba(FILE *f) {
  int l;
  fscanf(f, "%d", &l);
  return l;
  }

// funkcja przekazujaca liczbe graczowi
void wypisz(FILE *f, int k) {
  fprintf(f, "%d\n", k);
  // !
  fflush(f);
  }

// blad funkcji systemowej
void syserr (char *msg) {

  fprintf(stderr,"ERROR: %s (%d; %s)\n", msg, errno, strerror(errno));

  exit(2);
  }


// gdzie jestesmy
int gdzie;

// lacza do komunikacji z graczami
FILE *in[2], *out[2];

// tworzenie gracza (porozumiewajacego sie przez podane in i out)
int tworzGracza(FILE ** in, FILE ** out) {
  int pIn[2], pOut[2];
  
  if (pipe(pIn) == -1)
    syserr("pipe");

  if (pipe(pOut) == -1)
    syserr("pipe");

  int pid = fork();
  
  if(pid < 0)
    syserr("fork");
  
  if(pid == 0) {
    if(close(pIn[1])) syserr("close");
    if(close(0)) syserr("close");
    if(dup(pIn[0]) != 0) syserr("dup");
    if(close(pIn[0])) syserr("close");

    if(close(pOut[0])) syserr("close");
    if(close(1)) syserr("close");
    if(dup(pOut[1]) != 1) syserr("dup");
    if(close(pOut[1])) syserr("close");

    execl("./gracz", "./gracz", NULL);
    
    syserr("gracz");
    }
  
  if(pid > 0) {
    if(close(pIn[0])) syserr("close");
    *in = fdopen(pIn[1], "at");
    if(!in) syserr("fdopen");

    if(close(pOut[1])) syserr("close");
    *out = fdopen(pOut[0], "rt");
    if(!out) syserr("fdopen");
    return pid;
    }
  }

// czytanie danych
// uwaga: funkcja nie sprawdza poprawnosci
void dane() {
  N = liczba(stdin);
  M = liczba(stdin);
  for(int k=0; k<N; k++) {
    czyje[k] = liczba(stdin);
    ranga[k] = liczba(stdin);
    }
  for(int l=0; l<M; l++) {
    skad[l] = liczba(stdin);
    dokad[l] = liczba(stdin);
    }

  gdzie = liczba(stdin);
  }

// funkcja przekazujaca dane graczom (oprocz pozycji poczatkowej)
void przekazDane(FILE *in, int g) {
  wypisz(in, g);
  wypisz(in, N);
  wypisz(in, M);
  for(int k=0; k<N; k++) {
    wypisz(in, czyje[k]);
    wypisz(in, ranga[k]);
    }
  for(int l=0; l<M; l++) {
    wypisz(in, skad[l]);
    wypisz(in, dokad[l]);
    }
  }

// rozgrywka (po przekazaniu areny)
void rozgrywka() {
  // odwiedzone do tej pory pozycje
  int rozgrywka[maxN];

  // gdzie juz bylismy
  bool bylismy[maxN];

  // dlugosc dotychczasowej rozgrywki
  int dlugosc = 0;
  
  while(!bylismy[gdzie]) {

    bylismy[gdzie] = true;
    rozgrywka[dlugosc++] = gdzie;

    printf(
      "Jestesmy w: %d, ruch %s, ranga %d\n",
      gdzie, czyje[gdzie] ? "Adama" : "Ewy", ranga[gdzie]
      );
    
    for(int g=0; g<2; g++)
      wypisz(in[g], gdzie);
    
    int ruch = liczba(out[czyje[gdzie]]);
    
    // sprawdzamy poprawnosc ruchu
    bool ok = false;
    for(int l=0; l<M; l++)
      if(skad[l] == gdzie && dokad[l] == ruch)
        ok = true;

    if(!ok) {
      printf("Nieprawidlowy ruch do %d - gracz %d przegral\n", ruch, czyje[gdzie]);
      exit(0);
      }

    gdzie = ruch;
    }

  printf("Cykl (pozycja %d)\n", gdzie);
  
  int max = -1;
  do {
    dlugosc--;
    if(ranga[rozgrywka[dlugosc]] > max) max = ranga[rozgrywka[dlugosc]];
    }
  while(rozgrywka[dlugosc] != gdzie);

  printf(
    "Najwieksza liczba %d - gracz %d (%s) wygral\n",
    max, max & 1, max & 1 ? "Adam" : "Ewa"
    );
  
  for(int g=0; g<2; g++) {
    fclose(in[g]);
    fclose(out[g]);
    }
  }

int main() {

  dane();
  
  tworzGracza(&in[0], &out[0]);
  tworzGracza(&in[1], &out[1]);
  
  przekazDane(in[0], 0);
  przekazDane(in[1], 1);

  rozgrywka();  
  }

